# Waters Balance Model for Gatun Lake
# M. G. Castrellon | 23/02/2024

# Import libraries
import torch
import torch.nn as nn
from ann import *

# Define classes for the water balance

class RabeloMappingMlp(nn.Module):
    """
    This function maps from volumen to water balance of a lake.
    For the specific case of Gatun Lake, the architecture is based on
    the paper by Rabelo et al. (2014).
    """
    def __init__(self):
        super(RabeloMappingMlp, self).__init__()
        self.linear1 = nn.Linear(in_features=1, out_features=2)
        self.linear2 = nn.Linear(in_features=2, out_features=1)
    
    def forward(self, x):
        x = torch.relu(self.linear1(x))
        x = torch.relu(self.linear2(x))
        return x


class LakeWaterBalance(nn.Module):

    def __init__(self, V0, device, dt=1):
        super(LakeWaterBalance, self).__init__()
        self.device = device
        self.V0 = V0
        self.dt = dt
    
    def get_inflows(self):
        Qp = self.data['Precipitation']
        Qmd = self.data['Madden_Dam']
        Qtr = self.data['Tributaries']
        return Qp + Qmd + Qtr
    
    def get_outflows(self):
        Qe = self.data['Evaporation']
        Qgd = self.data['Gatun_Dam']
        Qwi = self.data['Water_Intakes']
        Qld = self.data['Lock_Total_Discharges']
        return Qe + Qgd + Qwi + Qld
    
    def calc_water_balance(self, data_dict):
        self.data = data_dict
        Qin = self.get_inflows()
        Qout = self.get_outflows()
        t = len(Qin)
        V = torch.zeros(t)
        V[0] = self.V0
        dt = self.dt
        for i in range(0, t-1):
            V[i+1] = V[i] + (Qin[i] - Qout[i])*dt
        return V.view(-1, 1).to(self.device)
    
    def forward(self, data_dict):
        v = self.calc_water_balance(data_dict)
        return v


class NeuralWaterBalance(nn.Module):

    def __init__(self, V0, device, n_hidden, dt=1, 
                 mapping='Rabelo'):
        super(NeuralWaterBalance, self).__init__()
        if mapping == 'Rabelo':
            self.mapping = RabeloMappingMlp()
        elif mapping == 'FNN':
            self.mapping = TwoLayerMlp(1, n_hidden, 1)
        self.model = LakeWaterBalance(V0, device, dt)
    
    def forward(self, data_dict):
        v = self.model(data_dict)
        h = self.mapping(v)
        return h