# Testing water balance for Gatun lake
# M. G. Castrellon | 5 March 2024

# %%

# Load libraries
import os
import sys
import glob
import torch
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

# Set the working directory
os.chdir("D:/SURFdrive/Projects/Differentiable_Model")

# Load custom modules
sys.path.append("src/model")
import water_balance as wb
import utilities as utils

# Define device
device = torch.device('cpu')

# %% 

# Load data

## Create list of files
folder_in = "./data/interim/hydromet"
glob_pattern = os.path.join(folder_in, "*.csv")
file_list = glob.glob(glob_pattern)

## Read files into a dataframe
df = pd.concat(
    (pd.read_csv(f) for f in file_list), 
    ignore_index=True
)

#%%

# Organize data

## Filter data by date
df['Date'] = pd.to_datetime(df['Date'])
df = df.loc[df['Date'] >= '2005-01-01', :]
df = df.loc[df['Date'] <= '2020-12-31', :]

## Filter data by label
df = df.loc[df['Label'].isin([
    'Precipitation', 
    'Evaporation', 
    'Madden_Dam', 
    'Gatun_Dam', 
    'Tributaries', 
    'Water_Intakes', 
    'Lock_Total_Discharges',
    'Water_Level'
]), :]

## Pivot table
df_pivot = df.pivot_table(
    index='Date', 
    columns='Label', 
    values='Value'
)

## Drop name of columns
df_pivot.columns.name = None

## Convert from mm/d to m3/d
lake_area = 436 * 10**6 # m2

df_pivot['Precipitation'] = \
    (df_pivot['Precipitation']/1000) * lake_area

df_pivot['Evaporation'] = \
    (df_pivot['Evaporation']/1000) * lake_area

# %%

# Run water balance model

## Transform data into tensor dictionary
data_dict = utils.df_to_tensor_dict(df_pivot, device)

## Define initial conditions
V0 = lake_area * 26 # m3

## Get observed water level
h_obs = data_dict['Water_Level']

## Define and run model
model = wb.LakeWaterBalance(V0=V0, device=device, dt=1)
volume = model(data_dict)
h_pred = volume / lake_area

## Plot results
plt.plot(h_obs.detach().numpy(), label='Observed')
plt.plot(h_pred.detach().numpy(), label='Predicted')
plt.legend()
plt.show()

# %%


